/**
 * Dasu Project
 *
 * File:         FindClass.java
 * RCS:          $Id: FindClass.java,v 1.0 Oct 11, 2012 9:57:55 AM mas939 Exp $
 * Description:  FindClass class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Oct 11, 2012 at 9:57:55 AM
 * Language:     Java
 * Package:      edu.northwestern.dasu.util
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2012, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.util;

import java.io.*;
import java.nio.*;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;


/**
 * @author Mario Sanchez &lt;msanchez@eecs.northwestern.edu&gt;
 *
 * The FindClass class ...
 */
public class FindClass {
    private static final int BUFFER_SIZE = 4096;
    private static final String[] JAR_EXTENSIONS = {".jar"};
    private static StringBuilder foundClasses = new StringBuilder();

    public static synchronized String findClass(String path,
                                 String classToLookFor,
                                 boolean recursive,
				 boolean reportVersion) 
                       throws Exception {

    	foundClasses = new StringBuilder();
    	
    	if (path == null) {
            return null;
        }
    	
        // normalize class name
        if (!classToLookFor.contains("/")) {
            classToLookFor = classToLookFor.replace('.', '/');
	    if (classToLookFor.endsWith("/class")) {
		classToLookFor = classToLookFor.substring(
				    0, 
				    classToLookFor.length() - 6
				 ) 
			       + ".class";
	    }
        }

        // normalize path, change \ to /
        path = path.replace(File.separatorChar, '/'); 
        File f = new File(path);
        if (f.isDirectory()) {
            findClassInDirectory(path, classToLookFor, recursive, reportVersion);
        }
        else if (f.isFile() && hasJarExtension(path)) {
            ZipInputStream in = new ZipInputStream(new FileInputStream(path));
            findClassInJar(in, classToLookFor, path, recursive, reportVersion);
            in.close();
        }
        else if (f.isFile() && path.contains(classToLookFor)) {
	    if (reportVersion) {
		InputStream in = new FileInputStream(f);
		String classVersion = getClassVersion(in);
		in.close();
		path += " (" + classVersion + ")";
	    }
            showPath(path);
	    	
        }
        
        return foundClasses.toString();

    }

    private static String getClassVersion(InputStream in) throws IOException {
	byte[] b = new byte[8];
	in.read(b);

	ByteBuffer bb = (ByteBuffer) ByteBuffer.allocate(8).put(b).rewind();

	int magic = bb.getInt();
	if (magic != 0xCafeBabe) { // not a Java class file
	    return "not a Java class";
	}

	return String.format("%2$d.%1$d", bb.getShort(), bb.getShort());
    }

    private static void findClassInJar(ZipInputStream in, 
                                      String classToLookFor,
                                      String path, 
                                      boolean recursive,
				      boolean reportVersion) 
                       throws Exception {
        ZipEntry next = in.getNextEntry();
        while (next != null) {
            if (!next.isDirectory()) {
                if (next.getName().contains(classToLookFor)) {
		    if (reportVersion) {
			String classVersion = getClassVersion(in);
			showPath(path 
			       + ": " 
			       + next.getName() 
			       + " (" + classVersion 
			       + ")");
		    }
		    else {
			showPath(path + File.separator + next.getName());
		    }
                }
                else if (recursive && hasJarExtension(next.getName())) {
                    ByteArrayOutputStream out = new ByteArrayOutputStream();
                    byte[] buffer = new byte[BUFFER_SIZE];
                    int n = in.read(buffer, 0, BUFFER_SIZE);
                    while (n >= 0) {  // read returns -1 if end of current entry is reached.
                        out.write(buffer, 0, n);
                        n = in.read(buffer, 0, BUFFER_SIZE);
                    }

                    ZipInputStream in2 = new ZipInputStream(
                                                new ByteArrayInputStream(
                                                    out.toByteArray()
                                                )
                                         );
                    findClassInJar(in2, 
                                   classToLookFor, 
                                   path + "/" + next.getName(), 
                                   recursive,
				   reportVersion);
                    in2.close();
                }
            }
            next = in.getNextEntry();
        }
    }

    private static void findClassInDirectory(String dir,
                                             String classToLookFor,
                                             boolean recursive,
					     boolean reportVersion)
                        throws Exception {
        File d = new File(dir);
        String[] files = d.list();
        for (int i = 0; i < files.length; i++) {
            String filePath = dir + "/" + files[i];
            File file = new File(filePath);
            if (file.isFile() && filePath.contains(classToLookFor)) {
                int idx = filePath.indexOf(classToLookFor);
		if (reportVersion) {
		    InputStream in = new FileInputStream(file);
		    String classVersion = getClassVersion(in);
		    in.close();
		    showPath(filePath.substring(0, idx) 
			   + ": " 
			   + filePath.substring(idx)
			   + " (" + classVersion + ")");
		}
		else {
		    showPath(filePath.substring(0, idx) + File.separator + filePath.substring(idx));
		}
            }
            else if (recursive) {
                if (file.isDirectory()) {
                    findClassInDirectory(filePath, classToLookFor, recursive, reportVersion);
                }
                else if (hasJarExtension(filePath)) {
                    ZipInputStream in = new ZipInputStream(new FileInputStream(file));
                    findClassInJar(in, classToLookFor, filePath, recursive, reportVersion);
                    in.close();
                }
            }
        }
    }

    private static boolean hasJarExtension(String fileName) {
        fileName = fileName.toLowerCase();
        for (int i = 0; i < JAR_EXTENSIONS.length; i++) {
            if (fileName.endsWith(JAR_EXTENSIONS[i])) {
                return true;
            }
        }
        return false;
    }

    private static void showPath(String p) {    	
    	foundClasses.append(p + "\n");
        //System.out.println(p);
    }

    public static void main(String[] args) throws Exception {
    	
    	args = new String[]{"/Users/mas939/Mario's Stuff/NORTHWESTERN/Dropbox/workspace/ISP Interference/testaz/testdasu/plugins/dasu/DasuUiRulesModules_2.0.jar",
    			"org.apache.zookeeper.client.FourLetterWordMain.class"};
    			//"edu.northwestern.dasu.dynamic.measurement.DasuScrapeModule.class"};
    			//"edu.northwestern.dasu.asn.NetworkAdminASNImpl"};
    	
        if (args.length < 2) {
            System.out.println("Usage: java FindClass [-v] <jarFileName> <classToLookFor>");
            System.exit(1);
        }
        
	boolean reportVersion = args[0].equals("-v");
        String path = reportVersion ? args[1] : args[0];
        String classToLookFor = reportVersion ? args[2] : args[1];
        System.out.println("found this " + findClass(path, classToLookFor, true, reportVersion));
    }
    
}